#include "common.h"
#include "sequence.h"
#include "../SequenceParser/SequenceParser.h"
#include "../resource.h"
#include "../util.h"
#include <algorithm>
#include "../timeobj.h"

CEventSequence::CEventSequence()
	:m_iLastSequenceIndex(-1)
{
}


CEventSequence::~CEventSequence()
{
	Clear();
}

//V[PX
void CEventSequence::Clear()
{
	m_iLastSequenceIndex=-1;
	for(unsigned int i=0;i<m_EventArray.size();i++){
		delete m_EventArray[i];
	}
	m_EventArray.clear();
}

//V[PX[hď
bool CEventSequence::Init()
{
	Clear();
	return m_CommandList.ReadFromFile();
}

//V[PX[hɏ
void CEventSequence::Reset()
{
	m_iLastSequenceIndex=-1;
	for(unsigned int i=0;i<m_EventArray.size();i++){
		m_EventArray[i]->Reset();
	}
}

//V[PX[h
bool CEventSequence::ReadFromFile(LPCTSTR lpFileName)
{
	Clear();
	ASSERT(lpFileName);
	if(!lpFileName)return false;

	std::vector<BYTE> Buffer;

	//t@Cǂݍ
	HANDLE hFile=CreateFile(lpFileName,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	if(INVALID_HANDLE_VALUE==hFile){
		UtilErrorMessage(NULL,CString(MAKEINTRESOURCE(IDS_ERROR_READ_EVEDNTFILE)));
		return false;
	}

	//4GB傫ȃt@CTCY͎擾łȂAIɂ͂̂悤ȑ傫ȃt@C͈Ȃ
	DWORD dwSize=GetFileSize(hFile,NULL);

	if(dwSize<=0){
		CloseHandle(hFile);
		UtilErrorMessage(NULL,CString(MAKEINTRESOURCE(IDS_ERROR_READ_EVEDNTFILE)));
		return false;
	}

	dwSize++;	//+1 is for '\0'
	Buffer.resize(dwSize);
	DWORD dwRead=0;
	BOOL Ret=ReadFile(hFile,&Buffer[0],dwSize-1,&dwRead,NULL);
	CloseHandle(hFile);
	if(!Ret||dwRead!=dwSize-1){
		UtilErrorMessage(NULL,CString(MAKEINTRESOURCE(IDS_ERROR_READ_EVEDNTFILE)));
		return false;
	}

	//I[ǉ
	Buffer[dwSize-1]='\0';

	//ʒusPʂŃCxgV[PXt@C
	DWORD dwIndex=0;
	CString Line;
	for(;dwIndex<dwSize;dwIndex++){
		TCHAR c=Buffer[dwIndex];
		if(c==_T('\n')||c==_T('\r')||c==_T('\0')){
			Line.Remove(_T('\n'));
			Line.Remove(_T('\r'));

			if(Line.IsEmpty())continue;

			DWORD dwTimeCode=0;
			std::string strEventName;
			if(!SequenceParser_GetEvent(Line,dwTimeCode,strEventName)){
				//p[XɎs
				UtilErrorMessage(NULL,CString(MAKEINTRESOURCE(IDS_ERROR_PARSE_EVENTFILE)));
				Clear();
				return false;
			}
			Line.Empty();
			//Cxgǉ
			LPCBYTE lpByteStream;
			DWORD dwLength;
			unsigned int uOutputPort;
			if(!m_CommandList.GetByteStream(strEventName,uOutputPort,lpByteStream,dwLength)){
				//oCgXg[擾Ɏs(R}h)
				CString msg(MAKEINTRESOURCE(IDS_ERROR_GETBYTESTREAM));
				msg+=_T("\r\n");
				msg+=strEventName.c_str();
				UtilErrorMessage(NULL,msg);
				Clear();
				return false;
			}
			if(uOutputPort>=CCommManager::GetInstance().GetPortCount()&&uOutputPort!=EXTERNAL_EXEC_COMMAND_PORT&&uOutputPort!=EXTERNAL_OPEN_COMMAND_PORT){
				//sȃ|[gIDԍ
				UtilErrorMessage(NULL,CString(MAKEINTRESOURCE(IDS_ERROR_SENDPORT)));
				Clear();
				return false;
			}
			m_EventArray.push_back(new CEventItem(dwTimeCode,strEventName.c_str(),uOutputPort,lpByteStream,dwLength));
		}
		else{
			Line+=c;
		}
	}

	//\[g
	SortByTime();
	return true;
}

bool CEventSequence::MakeFromCommandList()	//R}hXgV[PX쐬
{
	Clear();

	if(m_CommandList.IsEmpty())return false;

	COMMAND_MAP_ITERATOR ite,end;
	m_CommandList.GetIterator(ite,end);
	for(;ite!=end;ite++){
		//Cxgǉ
		LPCBYTE lpByteStream;
		DWORD dwLength;
		unsigned int uOutputPort;
		std::string strEventName((*ite).first);
		if(!m_CommandList.GetByteStream(strEventName,uOutputPort,lpByteStream,dwLength)){
			Clear();
			return false;
		}
		if(uOutputPort>=CCommManager::GetInstance().GetPortCount()&&uOutputPort!=EXTERNAL_EXEC_COMMAND_PORT&&uOutputPort!=EXTERNAL_OPEN_COMMAND_PORT){
			//sȃ|[gIDԍ
			UtilErrorMessage(NULL,CString(MAKEINTRESOURCE(IDS_ERROR_SENDPORT)));
			Clear();
			return false;
		}
		m_EventArray.push_back(new CEventItem(0,strEventName.c_str(),uOutputPort,lpByteStream,dwLength));
	}
	return true;
}

CEventItem* CEventSequence::GetEventSequence(int index)
{
	if(index<0||index>=(signed)m_EventArray.size())return NULL;

	return m_EventArray[index];
}

EVENT_STATUS CEventSequence::OnTimerCall()
{
	CTimeObject &TimeObject=CTimeObject::GetInstance();
	if(TimeObject.IsHalted()){
		//~͎sȂ
		return EVENT_NOT_EXECUTED;
	}

	const DWORD dwNow=TimeObject.GetTime();

	int m_iLastSequenceIndex_Backup=m_iLastSequenceIndex;

	//ɎsV[PXT
	for(int i=m_iLastSequenceIndex+1;i<(signed)m_EventArray.size();i++){
		CEventItem *lpEvent=m_EventArray[i];
		if(lpEvent->GetTimeCode()>dwNow){
			m_iLastSequenceIndex=i-1;
			if(m_iLastSequenceIndex==m_iLastSequenceIndex_Backup){
				return EVENT_NOT_EXECUTED;
			}
			else{
				return EVENT_EXECUTED;
			}
		}
		if(!lpEvent->SendEvent()){
			m_iLastSequenceIndex=i-1;
			return EVENT_ERROR;
		}
	}

	m_iLastSequenceIndex=i-1;
	return EVENT_FINISHED;	//I
}

//蓮s
bool CEventSequence::ManualExecute(int index)
{
	if(index<0||index>=(signed)m_EventArray.size())return false;

	CEventItem *lpEvent=m_EventArray[index];
	ASSERT(lpEvent);
	if(!lpEvent)return false;

	//ŏIsV[PX̃CfbNXXV
	m_iLastSequenceIndex=index;

	//ԂV[PX̃^CR[hɍ킹
	CTimeObject::GetInstance().Reset(lpEvent->GetTimeCode());

	//s
	return lpEvent->SendEvent();
}

//r֐IuWFNg
struct COMP{
	bool operator()(CEventItem* x, CEventItem* y)const{
		return (x->GetTimeCode() < y->GetTimeCode());
	}
};

//^CR[hŕבւ
void CEventSequence::SortByTime()
{
	COMP Comp;
	std::sort(m_EventArray.begin(),m_EventArray.end(),Comp);
}

